import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.Socket;
import java.net.UnknownHostException;



// the client end of a connection and the superclass of CGServer.
public class CGClient {
	
	// these fields need to be seen by the Server.
	
	// the size of the cycle in which the server waits for client connections
	// and in which the waiting party waits for incoming data.
	final public static int WAIT_MILLISECONDS = 500;

	// the port trough which the client calls in.
	final public static int PORT = 6979;
	
    // whether we are connected to an opponent or not.
    public boolean connected = false;
    
    // the output and input streams.
    public PrintWriter out = null;
    public BufferedReader in = null;

    
    
	// the client Socket.
    private Socket socket = null;

    
    
    // the constructor sets the parent game, the opponent's IP address
    // and initializes the protocol.
    private CrosswordGame game;
	private String opponentIP;
	public CGClient(final CrosswordGame g, final String ip) {
		game = g;
		opponentIP = ip;
	}
	
	
	
	// form a c2s connection, continuing an interrupted game.
	private boolean continueConnect() {
		
        try {
	        String inputLine;
	        
	        // ignore our letters, attempt to signal the server.
	        inputLine = in.readLine();
	        out.println(CGProtocol.CLIENT_REQUESTING_CONTINUE);
	        inputLine = in.readLine();
	        
	        // if permission to continue was granted by the server, start sending.
	        if ( CGProtocol.isContinueGranted(inputLine) ) {
		        return sendGame(true);
	        }
	        
	        // if the server also expresses a wish to continue,
	        // submit to it and start receiving.
	        else if ( CGProtocol.isSPC(inputLine) ) {
	        	out.println(CGProtocol.CLIENT_ACKNOWLEDGES_SPC);
	        	return receiveGame();
	        }
	        
	        // if neither a permission to continue nor an overriding continue wish 
	        // was received, we're not talking the same language with the server. 
	        else {
	        	System.err.println("DEBUG: CGClient: Asiakkaan tekem jatkopyynt eponnistui.");
	        	return false;
	        }
        }

        // catch unexpected connection breakdowns.
        catch (IOException e) {
        	//System.err.println("DEBUG: CGClient: Palvelinyhteys katkesi odottamattomasti.");
        	return false;
		}
	}

	
	
    // provide meta methods for CGConnection to access I/O streams.
	public String readLine() throws IOException {
		if (in == null)	throw new IOException();
		if (!connected) throw new IOException();
		return in.readLine();
	}
	public void writeLine(String s) throws IOException {
		if (out == null)throw new IOException();
		if (!connected) throw new IOException();
		out.println(s);
	}

    // the method for shutting down the client connection.
	// CGConnection needs to see this.
	public void initialize() {
		try {
			if (in != null) {
				in.close();
			}
			if (out != null) {
				out.close();
			}
			if (socket != null) {
				socket.close();
			}
		}
        catch (IOException e) {
        }
        
        connected = false;
	}

	// send an encoded move out to the opponent.
	// CGConnection needs to see this.
    // TODO: threading?
	public boolean sendMove(String move) {
        try {
	        out.println(move);
	        if ( !CGProtocol.isMoveReceived(readLine()) )	return false;
	        else									        return true;
        }
        
        // catch unexpected connection breakdowns.
        catch (IOException e) {
        	//System.err.println("DEBUG: CGClient: Palvelinyhteys katkesi odottamattomasti.");
        	return false;
		}
	}
	
	
	// form a c2s connection.
	// needs to be seen by CGConnection.
	// is overridden in CGServer.
	public boolean connect() {	return connect(false);	}
	public boolean connect(boolean continueGame) {
        try {
            socket = new Socket(opponentIP, PORT);
    		socket.setSoTimeout(WAIT_MILLISECONDS);
        }
        
		// TODO: a nice way of announcing UnknownHostException.
        catch (UnknownHostException e) {
        	//System.err.println("DEBUG: CGClient: Tuntematon osoite: " + opponentIP + ".");
        	return false;
        }
        catch (IOException e) {
        	//System.err.println("DEBUG: CGClient: Palvelinyhteyden muodostaminen eponnistui.");
        	return false;
        }
        // the Socket is now established.
        
    	// establish input and output streams.
        try {
            out = new PrintWriter(socket.getOutputStream(), true);
            in = new BufferedReader(new InputStreamReader(socket.getInputStream()));
        }
        
        catch (IOException e) {
        	System.err.println("DEBUG: CGClient: Tietovirtojen kytkent eponnistui.");
        	return false;
        }
        // the input and output streams are now established.

        // if attempting to continue an interrupted game, signal the server.
        // otherwise read game initialization data from the server.
        if (continueGame)	return continueConnect();
        else				return receiveGame();
	}

    // read game initialization data from the opponent.
	// CGServer needs to see this.
	public boolean receiveGame() {
        try {
	        String inputLine;
	        
	        // receive initialization data, sending out acknowledgment
	        // when data is successfully applied.
	        
	        // our slots.
	        inputLine = in.readLine();
	        if ( !game.playerSlots.set(inputLine) )				return false;
	        out.println(CGProtocol.LETTERSLOTS_RECEIVED);
	
	        // opponent's slots.
	        inputLine = in.readLine();
	        boolean opssuccess = game.opponentSlots.set(inputLine);
	    	game.opponentSlots.disable();
	        if ( !opssuccess )			return false;
	        out.println(CGProtocol.LETTERSLOTS_RECEIVED);

	        // our score.
	        inputLine = in.readLine();
	        if ( !game.scoreTable.setPlayerScore(inputLine) )	return false;
	        out.println(CGProtocol.SCORE_RECEIVED);

	        // opponent's score.
	        inputLine = in.readLine();
	        if ( !game.scoreTable.setOpponentScore(inputLine) )	return false;
	        out.println(CGProtocol.SCORE_RECEIVED);

	        // the crossword grid.
	        inputLine = in.readLine();
	        if ( !game.crossword.set(inputLine) )				return false;
	        out.println(CGProtocol.CROSSWORD_RECEIVED);
	        
	        // the Alphabet.
	        inputLine = in.readLine();
	        if ( !game.alphabet.set(inputLine) )				return false;
	        out.println(CGProtocol.ALPHABET_RECEIVED);
	        
	        // update alphabet counter.
	        game.scoreTable.setAlphabetCounter();

	        connected = true;
	        return true;
        }

        // catch unexpected connection breakdowns.
        catch (IOException e) {
        	initialize();
        	return false;
        }
	}
	
    // send game initialization data to opponent.
	// CGServer needs to see this.
	public boolean sendGame() {	return sendGame(false);	}
	public boolean sendGame(boolean continueRequested) {
        try {
	        String inputLine;

	        // send initialization data, waiting for acknowledgments
	        // when data is successfully applied.
	        
	        // opponent's slots.
	        out.println(game.opponentSlots.toString());
	        inputLine = in.readLine();
        	if ( CGProtocol.isClientRequestingContinue(inputLine) ) {
        		
        		// a conflict of continuation requests.
        		if (continueRequested) {
        			
        			out.println(CGProtocol.SERVER_PROVIDES_CONTINUE);
        	        inputLine = in.readLine();
        	        
        			if ( !CGProtocol.isClientAcknowledgesSPC(inputLine) )	return false;
        			
    	        	// the client submits to the server's will: we're clear to try again.
        			return sendGame(continueRequested);
        		}
        		
        		
       			// receive game initialization data from the client.
           		else {
        			out.println(CGProtocol.CONTINUE_GRANTED);
        			return receiveGame();
        		}
        	}
        	
   			// receive game initialization data from the server.
        	else if ( CGProtocol.isServerRequestingContinue(inputLine) ) {
    			out.println(CGProtocol.CONTINUE_GRANTED);
    			return receiveGame();
    		}
        	
        	else if ( !CGProtocol.isSlotsReceived(inputLine) )		return false;
	
	        // our slots.
	        out.println(game.playerSlots.toString());
	        inputLine = in.readLine();
	        if ( !CGProtocol.isSlotsReceived(inputLine) )			return false;

	        // opponent's score.
	        out.println(game.scoreTable.opponentScoreToString());
	        inputLine = in.readLine();
	        if ( !CGProtocol.isScoreReceived(inputLine) )			return false;
	        
	        // our score.
	        out.println(game.scoreTable.playerScoreToString());
	        inputLine = in.readLine();
	        if ( !CGProtocol.isScoreReceived(inputLine) )        	return false;
	        
	        // the crossword.
	        out.println(game.crossword.toString());
	        inputLine = in.readLine();
	        if ( !CGProtocol.isCrosswordReceived(inputLine) )		return false;
	        
	        // the Alphabet.
	        out.println(game.alphabet.toString());
	        inputLine = in.readLine();
	        if ( !CGProtocol.isAlphabetReceived(inputLine) )       	return false;

	        connected = true;
	        return true;
		}

        // catch unexpected connection breakdowns.
        catch (IOException e) {
        	return false;
        }
	}
}
