import java.util.Vector;

import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Shell;



// the crossword grid, onto which the letters are placed. 
public class CGCrossword extends Composite {
	
	// word cancel status reports.
	final private static String CANCEL_CANCELLED_LETTERS = "Siivottiin pydlt kirjaimet";
	final private static String CANCEL_NO_UNFIXED = "Pydll ei ole siivottavia kirjaimia.";
	
	// the grid's dimensions.
	// the board is square-shaped, so the numbers are equal.
	// CrosswordGameSettings needs to know about these.
	final public static int COLUMNS = 15;
	final public static int ROWS = COLUMNS;
	
	// the coordinates of the centermost cell.
	final private static int CENTER_X = 7;
	final private static int CENTER_Y = CENTER_X;
	
	// an instance of the specialized class for word searching and other advanced algorithms.
	final private CGSearcher searcher;
	
	// the grid itself is an array of CGLetterSlots.
	final private CGCell b[][] = new CGCell[COLUMNS][ROWS];

	// the undo buffer.
	final private Vector<CGCell> undoBuffer = new Vector<CGCell>();
	
	
	
	// the constructor sets the parent game. the other parameter is the parent window.
	final private CrosswordGame game;
	public CGCrossword(final Shell shell, final CrosswordGame g) {
		
		super(shell, SWT.NONE);
		game = g;
		
		// an instance of the specialized class for word searching
		// and other advanced algorithms.
		searcher = new CGSearcher(this, game);

		// layout is based on the dimension set above.
		GridLayout gl = new GridLayout(COLUMNS, true);
		GridData gd = new GridData();
		gd.widthHint = 30;
		this.setLayout(gl);
		this.setLayoutData(gd);

    	// fill the grid with CGCells.
		for (int i=0; i<ROWS; i++) {
			for (int j=0; j<COLUMNS; j++) {
				b[j][i] = new CGCell(j, i, this, game);
			}
		}
		
		// do a grid-wide initialization just to be sure all cells are properly reset
		// and to set the center cell.
		initialize();
	}
	
	// clean up all letters from the board, set the center cell.
	// NOTE: the letters aren't moved back to LetterSets
	// but wasted, so this method should not be called
	// without also re-setting LetterSets and the Alphabet
	// immediately afterwards.
	public void initialize() {
		for (int i=0; i<ROWS; i++) {
			for (int j=0; j<COLUMNS; j++) {
				b[j][i].initialize();
			}
		}
		b[CENTER_X][CENTER_Y].setAsCenter();
	}
	
	// disable/enable the grid.
	// these need to be seen by CrosswordGame.
	public void disable() {
		for (int i=0; i<ROWS; i++) {
			for (int j=0; j<COLUMNS; j++) {
				b[j][i].disable();
			}
		}
	}
	public void enable() {
		for (int i=0; i<ROWS; i++) {
			for (int j=0; j<COLUMNS; j++) {
				b[j][i].enable();
			}
		}
	}
	
	// find the x coordinate corresponding to a given CGNewCell.
	// returns -1
	// if given CGNewCell is not in the grid.
	public int xOf(final CGCell s) {
		for (int i=0; i<ROWS; i++) {
			for (int j=0; j<COLUMNS; j++) {
				if ( b[j][i].equals(s)) {
					return j;
				}
			}
		}
		return -1;
	}
	
	// find the y coordinate corresponding to a given Button.
	// returns -1
	// if given Button is not found.
	public int yOf(final CGCell s) {
		for (int i=0; i<ROWS; i++) {
			for (int j=0; j<COLUMNS; j++) {
				if ( b[j][i].equals(s)) {
					return i;
				}
			}
		}
		return -1;
	}

	// whether given CGNewCell is the center cell.
	// the queue needs to see this.
	public boolean isCenter(CGCell s) {
		if ( (xOf(s) == CENTER_X) && (yOf(s) == CENTER_Y) )	return true;
		else												return false;
	}
	
	// grab a cell in a given position. returns null if coordinates are invalid.
	public CGCell getCrosswordCell(int x, int y) {
		if ( (x < COLUMNS) && !(x < 0) && (y < ROWS) && !(y < 0) )	return b[x][y];
		else														return null;
	}

	// fix & push into the undo buffer all unfixed letters currently on the table.
	// returns the number of letters fixed.
	public int fixWords() {
		
		// clear the undo buffer.
		undoBuffer.removeAllElements();
		
		int nFixed = 0;
		for (int i=0; i<ROWS; i++) {
			for (int j=0; j<COLUMNS; j++) {
				if (b[j][i].fixed) {
					b[j][i].setAsOld();
				}
				else if (!b[j][i].isEmpty()) {
					if (b[j][i].fix()) {
						b[j][i].setAsNew();
						nFixed++;
						undoBuffer.add(b[j][i]);
					}
				}
			}
		}

		return nFixed;
	}

	// undo opponent's last move.
	public void undoOpponent() {
		
		// a sanity check.
		if ( !(undoBuffer.size() > 0) )	return;
		
		for (int i=0; i<undoBuffer.size(); i++) {
			CGLetterSlot empty = game.opponentSlots.nextEmpty();
			empty.setLetter(new CGLetter(undoBuffer.elementAt(i).getLetter().toString()));
			undoBuffer.elementAt(i).initialize();
		}
	}
	
	// undo player's last move.
	public void undoPlayer() {
		
		// a sanity check.
		if ( !(undoBuffer.size() > 0) )	return;
		for (int i=0; i<undoBuffer.size(); i++) {
			CGLetterSlot empty = game.playerSlots.nextEmpty();
			empty.setLetter(new CGLetter(undoBuffer.elementAt(i).getLetter().toString()));
			undoBuffer.elementAt(i).initialize();
		}
	}
	
	// clean up all unfixed letters from the board, back to local player's LetterSet.
	public void cancelWords() {	cancelWords(false); }
	public void cancelWords(boolean quiet) {
		game.queue.unset();
		String s = "";
		for (int i=0; i<ROWS; i++) {
			for (int j=0; j<COLUMNS; j++) {
				if (!b[j][i].isEmpty() && !b[j][i].fixed) {
					
					// find the next empty letter slot from local player's LetterSet.
					CGLetterSlot empty = game.playerSlots.nextEmpty();
					
					// if there aren't any empty slots, fail quietly.
					// this can and will occur when connections are lost abruptly;
					// any letters laid onto the board by the opponent will linger
					// when the game is enabled.
					if (empty == null)	return;
					
					// make a note of the swapped letter for status update.
					s = s + b[j][i].getLetter().toString() + ", ";

					game.queue.set(b[j][i]);
					game.queue.swap(empty);
				}
			}
		}
		
		// update status, unless denied doing so.
		if (!quiet) {
			if (s.equals("")) {
				game.setStatus(CANCEL_NO_UNFIXED);
			}
			else {
				String sFinal = CANCEL_CANCELLED_LETTERS + " " + s;
				sFinal = sFinal.substring(0, sFinal.length()-2);
				sFinal = sFinal + ".";
				game.setStatus(sFinal);
			}
		}
	}
	
	// clean up all unfixed letters from board
	// back to opponent's LetterSet.
	public void cancelOpponentWords() {
		for (int i=0; i<ROWS; i++) {
			for (int j=0; j<COLUMNS; j++) {
				if (!b[j][i].isEmpty() && !b[j][i].fixed) {

					// push the cell contents to the queue.
					game.queue.set(b[j][i]);
					
					// find the next empty letter slot from opponent's LetterSet.
					CGLetterSlot empty = game.opponentSlots.nextEmpty();
					
					// swap the cell contents with the empty slot contents.
					game.queue.swap(empty);
				}
			}
		}
	}

	// return the board encoded in a String, for transmitting to opponent.
	public String toString() {
		
		// each cell is described by two letters,
		// the first of which is the letter held by the cell,
		// and the second a String-encoded (un)fixed status,
		// where F = fixed and !F = unfixed.
		String s = "";
		
		for (int i=0; i<ROWS; i++) {
			for (int j=0; j<COLUMNS; j++) {
				if (b[j][i].isEmpty())	s = s + CGProtocol.CROSSWORD_EMPTY_LETTER;
				else					s = s + b[j][i].getLetter().toString();
				if (b[j][i].fixed)		s = s + CGProtocol.CROSSWORD_FIXED_LETTER;
				else					s = s + CGProtocol.CROSSWORD_UNFIXED_LETTER;
			}
		}
		return s;
	}
	
	// set the board according to given String-encoded data.
	public boolean set(String sCrossword) {
		
		// a simple test for validity.
		if (sCrossword.length() != ROWS*COLUMNS*2)	return false;
		
		// clear the board.
		initialize();
		
		// interpret the data.
		for (int i=0; i<ROWS; i++) {
			for (int j=0; j<COLUMNS; j++) {
				String letter = sCrossword.substring(2*i*ROWS+2*j, 2*i*ROWS+2*j+1);
				String fixed = sCrossword.substring(2*i*ROWS+2*j+1, 2*i*ROWS+2*j+2);
				if (!letter.equals(CGProtocol.CROSSWORD_EMPTY_LETTER))	b[j][i].setLetter(new CGLetter(letter));
				if ( fixed.equals(CGProtocol.CROSSWORD_FIXED_LETTER) )	b[j][i].fix();
				else if ( !letter.equals(CGProtocol.CROSSWORD_EMPTY_LETTER))	b[j][i].highlight();
			}
		}
		
		return true;
	}
	
	// count letters on the board.
	// CrosswordGame needs to see this.
	public int countLetters() {
		int count = 0;
		for (int i=0; i<ROWS; i++) {
			for (int j=0; j<COLUMNS; j++) {
				if (!b[j][i].isEmpty()) {
					count++;
				}
			}
		}
		return count;
	}
	
	// count unfixed letters on the board.
	// (this is a dumb evaluation of last move's score.)
	// CGActionButtonSet needs to see this.
	public int countNew() {
		int count = 0;
		for (int i=0; i<ROWS; i++) {
			for (int j=0; j<COLUMNS; j++) {
				if (!b[j][i].isEmpty() && !b[j][i].fixed) {
					count++;
				}
			}
		}
		return count;
	}
	
	// a 'smart' scoring algorithm:
	// give points based on the word inserted by the player. 
	// CGActionButtonSet needs to see this.
	public int smartScore() {
		return searcher.getInsertedWord().length();
	}
	
	// get the word inserted by the player.
	// CGActionButtonSet needs to see this.
	public String getInsertedWord() {
		return searcher.getInsertedWord();
	}
}
