import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.text.Collator;
import java.util.Collections;
import java.util.Vector;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.events.VerifyEvent;
import org.eclipse.swt.events.VerifyListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Dialog;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.List;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;

// the dictionary.
public class CGDictionary {
	
	// the dictionary file name,
	// the maximum number of words,
	// the maximum and minimum lengths of one word.
	// CGDictionaryDialog needs to know the limits.
	final private static String DICTIONARY_FILENAME	= "dictionary.txt";
	final public static int	DICTIONARY_MAXSIZE		= 47000;
	final public static int	DICTIONARY_WORD_MAXSIZE	= CGCrossword.COLUMNS;
	final public static int DICTIONARY_WORD_MINSIZE = 2;
	
	// the dictionary is a Vector of Strings.
	private Vector<String> words;

	// the dialog needs the main window for setup,
	// so the opening method has to be able to reach it.
	// other classes need not know about it.
	final private Shell shell;
	
	// the constructor initially sets an empty dictionary.
	// loading a saved dictionary needs to be done separately
	// by invoking load().
	public CGDictionary(final Shell s) {
		words = new Vector<String>();
		shell = s;
	}
	
	// dictionary Vector encapsulation.
	public int size() {
		return words.size();
	}
	public String wordAt(final int index) {
		if (index < 0)	return null;
		if (index > words.size()-1)	return null;
		return words.elementAt(index);
	}
	public boolean contains(String word) {
		return words.contains(word);
	}
	public int indexOf(String word) {
		return words.indexOf(word);
	}
	public boolean remove(String word) {
		return words.remove(word);
	}
	public boolean add(String word) {
		
		// if the word already is in the dictionary,
		// skip insertion silently.
		if (words.contains(word)) {
			return false;
		}
		
		int i = getIndex(word);
		words.insertElementAt(word, i);
		return true;
	}
	
	// return the index into which a word must be inserted
	// in order to maintain alphabetical order.
	private int getIndex(String word) {
		
		// go through the dictionary, comparing the alphabetical order
		// of each word with the given word.
		// return the index of the one that is
		// alphabetically later than given word.
		int i;
		for (i=0; i<words.size(); i++) {
			if (Collator.getInstance().compare(word, words.elementAt(i)) < 0)
				break;
		}
		return i;
	}
	
	// returns the index of closest match to given word in the dictionary.
	public int closestMatch(String word) {
		int index = getIndex(word);
		if (contains(word))	index = indexOf(word);
		else	index = (index < size()) ? index : index-1;
		return index;
	}
	
	// dictionary saver.
	// returns TRUE on success, FALSE on failure (an IOException).
	public boolean save() {
	    try {
	    	
	    	// open the file for writing.
	    	final PrintStream out = new PrintStream(new FileOutputStream(DICTIONARY_FILENAME));

	    	// sort the dictionary alphabetically.
	    	Collections.sort(words, Collator.getInstance());
	    	
	    	// write the dictionary.
	    	for (int i=0; i<words.size(); i++)	out.println(words.elementAt(i));

	    	// close the file.
	    	out.close();
	
	    	// report success.
	    	return true;
	      
	    }

	    // if unable to write to file, report failure.
	    catch (IOException ex) {
	    	return false;
	    }
	}

	// validate a String for inclusion in the dictionary.
	// CGDictionaryDialog and CGSubmitDialog need to know about this.
	public boolean isValidWord(String word) {
		if (word == null)								return false;
		if (word.length() < DICTIONARY_WORD_MINSIZE)	return false;
		
		for (int i=0; i < word.length(); i++)
			if (!Character.isLetter(word.charAt(i)))	return false;

		return true;
	}
	
	// dictionary loader.
	// returns TRUE on success, FALSE on failure.
	// success is defined as having read at least one word successfully.
	public boolean load() {
		try {
			
			// open the file.
			final BufferedReader in = new BufferedReader(new FileReader(DICTIONARY_FILENAME));
			
			// initialize the dictionary.
			words.removeAllElements();
			
			// number of words read.
			int readCount = 0;
			
			// read until EOF.
			String word = in.readLine();
			while (word != null) {
				
				// if what was read is a valid word,
				// push it into the dictionary.
				if (isValidWord(word)) {
					words.addElement(word);
					readCount++;
				}
				
				// read only a preset amount of items.
				if ( !(readCount < DICTIONARY_MAXSIZE) )	break;
				
				// read until the next newline.
				word = in.readLine();
			}

			// close the file.
			in.close();

	    	// sort the dictionary alphabetically.
	    	Collections.sort(words, Collator.getInstance());

			// return TRUE if any words were read.
			return (words.size() > 0);
		}
		
		// fail silently on read errors.
		catch (FileNotFoundException ex) {
			return false;
		}
		catch (IOException ex) {
			return false;
		}
	}
	
	// "opening the dictionary" means opening a CGDictionaryDialog.
	public void open() {
		open("");
	}
	public void open(String word) {
		if (word == null)	new CGDictionaryDialog(shell).open("");
		else				new CGDictionaryDialog(shell).open(word);
	}
	
	
	
	// the dictionary is visualized in a dialog.
	private class CGDictionaryDialog	extends Dialog
				implements VerifyListener, ModifyListener, SelectionListener {
		
		final private static String TITLE = "Sanasto";

		// the dialog introduction text.
		final private static String PROMPT =
			"Alla olevasta sanastosta voit tarkistaa, onko miettimsi sanaa hyvksytty " +
			"aikaisemmissa peleiss. Tarvittaessa voit mys list (+) ja poistaa (-) sanoja.";
		
		// EditTools' tooltips.
		final private static String ADD_TOOLTIP		= "Lis sytetyn sanastoon.";
		final private static String DELETE_TOOLTIP	= "Poistaa valitun sanan sanastosta.";
		
		// 'Dictionary is full' message box title and message.
		final private static String DICTIONARY_FULL_TITLE	= "Ilmoitus: sanasto on tynn";
		final private static String DICTIONARY_FULL_MESSAGE	=
			"Uusia sanoja ei voida list, koska\n" +
			"sanaston koko on saavuttanut sille\n" +
			"asetetun ylrajan (" + CGDictionary.DICTIONARY_MAXSIZE + " sanaa).";

		
		
		// the dialog window. 
		private Shell shell;

		// the visible word List.
		private List list;

		// a text field for searching and adding words.
		private Text word;
		
		// Buttons for adding and deleting words.
		private Button add;
		private Button delete;
		
		// whether any changes have been made to the dictionary.
		private boolean changesMade;
		
		
		
		// the constructor.
		public CGDictionaryDialog(final Shell parent) {
			super (parent, SWT.NONE);
			changesMade = false;
		}

		
		// open()'s parameter is a String based on which the initial selection is done.
		public void open() {	open("");	}
		public void open(String initialSelection) {
			
			initialSelection = initialSelection.toUpperCase();
			
			Shell parent = getParent();
			shell = new Shell(parent, SWT.DIALOG_TRIM | SWT.APPLICATION_MODAL);
			shell.setLayout(new GridLayout(3, false));
			shell.setSize(200, 500);
			shell.setText(TITLE);

			// a prompt, introducing the usage.
			Label prompt = new Label(shell, SWT.WRAP);
			prompt.setText(PROMPT);
			GridData gdPrompt = new GridData(GridData.FILL_HORIZONTAL);
			gdPrompt.horizontalSpan = 3;
			prompt.setLayoutData(gdPrompt);
			
			// a text field for searching and inserting new words.
			word = new Text(shell, SWT.SINGLE | SWT.BORDER);
			word.setTextLimit(CGDictionary.DICTIONARY_WORD_MAXSIZE);
			word.setText(initialSelection);
			word.addVerifyListener(this);
			word.addModifyListener(this);
			GridData gdText = new GridData();
			gdText.widthHint = 120;
			word.setLayoutData(gdText);

			// a Button for adding words.
			add = new Button(shell, SWT.PUSH);
			add.setText("+");
			add.addSelectionListener(this);
			add.setToolTipText(ADD_TOOLTIP);
			
			// a Button for deleting words.
			delete = new Button(shell, SWT.PUSH);
			delete.setText("-");
			delete.addSelectionListener(this);
			delete.setToolTipText(DELETE_TOOLTIP);

			GridData gdButtons = new GridData();
			gdButtons.widthHint = 21;
			add.setLayoutData(gdButtons);
			delete.setLayoutData(gdButtons);

			// the visible dictionary is a List of words.
			GridData fill = new GridData(GridData.FILL_HORIZONTAL);
			fill.horizontalSpan = 3;
			fill.heightHint = 350;
			list = new List(shell, SWT.BORDER | SWT.V_SCROLL);
			list.addSelectionListener(this);
			list.setLayoutData(fill);
			for (int i=0; i<size(); i++)	list.add(wordAt(i));

			// set the initial selection.
			if (!initialSelection.equals(""))	selectClosestMatch(initialSelection);
			
			shell.open();
			Display display = parent.getDisplay();
			while (!shell.isDisposed()) {
				if (!display.readAndDispatch()) display.sleep();
			}
			
			// when the window is closed, save changes made to the dictionary, if any.
			if (changesMade)	save();
		}

		// find closest match to given word from the dictionary.
		private void selectClosestMatch(String w) {
			list.removeSelectionListener(this);
			
			list.select(closestMatch(w));
			list.showSelection();
			
			if (isValidWord(w))	add.setEnabled(true);
			else				add.setEnabled(false);
			
			list.addSelectionListener(this);
		}

		public void widgetDefaultSelected (SelectionEvent event) {	widgetSelected(event);	}
		public void widgetSelected (SelectionEvent event) {
			
			// selections made from the list.
			if (event.widget == list) {
				
				// modify listeners need to be disabled to be able to change the text.
				word.removeVerifyListener(this);
				word.removeModifyListener(this);
				
				word.setText(list.getSelection()[0]);
				
				word.addVerifyListener(this);
				word.addModifyListener(this);
				
			}
			
			// 'Add' button selected.
			else if (event.widget == add) {
				
				// if dictionary is not full,
				// insert the new word.
				// this assumes that the word is valid.
				if (size() < CGDictionary.DICTIONARY_MAXSIZE) {
					
					String w = word.getText();
					int i = indexOf(w);
					if (i < 0) {
						add(w);
						i = indexOf(w);
						list.add(w, i);
						changesMade = true;
					}
					// if the word already is in the dictionary,
					// skip insertion silently.
					
					list.removeSelectionListener(this);
					list.select(i);
					list.addSelectionListener(this);
					word.setFocus();
					return;
				}
				
				// if dictionary is full,
				// notify the player.
				else {
					MessageBox dictionaryFull = new MessageBox(shell, SWT.OK);
					dictionaryFull.setText(DICTIONARY_FULL_TITLE);
					dictionaryFull.setMessage(DICTIONARY_FULL_MESSAGE);
					dictionaryFull.open();
				}
			}

			// 'Delete' button selected.
			else if (event.widget == delete) {
				
				// grab the entered word and current list selection.
				String w = word.getText();
				int i = list.getSelectionIndex();
				
				// if the word is in the dictionary, remove it.
				if (contains(w)) {
					remove(w);
					changesMade = true;
					
					// select the next item on the list 
					if (i == -1)	i = 0;
					else			list.remove(i);
					i = (i < size()) ? i : size()-1;
					
					list.removeSelectionListener(this);
					word.removeVerifyListener(this);
					word.removeModifyListener(this);
					list.select(i);
					word.setText(list.getItem(i));
					list.addSelectionListener(this);
					word.addVerifyListener(this);
					word.addModifyListener(this);
					
				}
				
				// if the word isn't in the dictionary,
				// remove it from the Text widget.
				else {
					word.removeVerifyListener(this);
					word.removeModifyListener(this);
					word.setText("");
					word.addVerifyListener(this);
					word.addModifyListener(this);
				}
				
				if (isValidWord(word.getText()))	add.setEnabled(true);
				else								add.setEnabled(false);
			}
			word.setFocus();
		}
		
		// spring to action when a character is typed.
		public void verifyText(VerifyEvent event) {
			
			// only allow backspace, delete and letters. convert to uppercase on-the-fly.
			// http://www.java2s.com/Tutorial/Java/0280__SWT/Allowbackspaceanddelete.htm
			event.doit = false;
			char c = event.character;
			if (c == '\u0008' || c == '\u007F') event.doit = true;
			else if (Character.isLetter(c))	{
				event.text = event.text.toUpperCase();
				event.doit = true;
			}
		}

		// spring to action after text is changed:
		// find closest match from the dictionary and select it.
		public void modifyText(ModifyEvent event) {	selectClosestMatch(word.getText());	}
	}
}
