import java.util.Vector;

import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;



public class CGLetterSlots extends Composite {
	
	// the maximum amount of slots in a CGLetterSlots.
	// CGAlphabet and CGScoretable need to see this.
	final public static int MAXSLOTS = 7;

	// the actual set of letter slots is a Vector.
	final private Vector<CGLetterSlot> slots = new Vector<CGLetterSlot>();

	// the undo buffer.
	final private Vector<CGLetterSlot> undoBuffer = new Vector<CGLetterSlot>();
	
	// the constructor takes as parameters the title string, the parent window and game.
	final private CrosswordGame game;
	public CGLetterSlots(final String o, final Shell shell, final CrosswordGame g) {

		super(shell, SWT.BORDER);
		game = g;
		GridLayout gl = new GridLayout();
		gl.numColumns = MAXSLOTS;
		this.setLayout(gl);

		// set the title unless an empty title was requested.
		if (o != null) {
			if (!o.equals("")) {
				final Label title = new Label(this, SWT.CENTER);
				title.setText(o);
				GridData gd = new GridData(GridData.FILL_HORIZONTAL);
				gd.horizontalSpan = MAXSLOTS;
				title.setLayoutData(gd);
			}
		}
		
        // fill the Vector with slots.
		for (int i=0; i<MAXSLOTS; i++) {
			slots.addElement(new CGLetterSlot(this, game));
			slots.elementAt(i).setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true, 1, 1));
		}
	}
	
	// pick new letters for the slots.
	// the parent game needs to know about this.
	public void initialize() {
		for (int i=0; i<slots.size(); i++) {
			slots.elementAt(i).setLetter(new CGLetter(game.alphabet.pick()));
		}
	}
	
	// wrappers for the slots Vectors' methods.
	public CGLetterSlot elementAt(int i) {
		return slots.elementAt(i);
	}
	public boolean contains(CGLetterSlot s) {
		return slots.contains(s);
	}
	public int size() {
		return slots.size();
	}
	// disable/enable all slots.
	public void disable() {
		for (int i=0; i<slots.size(); i++) {
			slots.elementAt(i).disable();
		}
	}
	public void enable() {
		for (int i=0; i<slots.size(); i++) {
			slots.elementAt(i).enable();
		}
	}
	
	// get the index of given slot.
	// returns -1 if given slot is not in the set.
	public int indexOf(CGLetterSlot s) {
		if (slots.contains(s))	return slots.indexOf(s);
		else					return -1;
	}
	
	// find the next empty letter slot.
	// returns null when no empty slots are found.
	public CGLetterSlot nextEmpty() {
		for (int i=0; i<slots.size(); i++)
			if (slots.elementAt(i).isEmpty()) return slots.elementAt(i);
			
		return null;
	}
	
	// fill all empty slots with letters from the Alphabet.
	// push filled slots into the undo buffer.
	public void fill() {
		CGLetterSlot empty;
		String letter;

		// clear the undo buffer.
		undoBuffer.removeAllElements();

		// find the next empty slot.
		while ( (empty = nextEmpty()) != null ) {
			
			// pick next letter.
			// if no more letters are available, stop filling.
			if ( (letter = game.alphabet.pick()) == CGAlphabet.NOLETTER)	break;

			// fill the slot, push it into the undo buffer.
			else {
				empty.setLetter(new CGLetter(letter));
				undoBuffer.add(empty);
			}
		}
	}
	
	// undo a fill.
	public void undo() {
		
		// the order shouldn't really matter, since it'll be the same at both ends.
		for (int i=0; i<undoBuffer.size(); i++) {
			CGLetterSlot letter = undoBuffer.elementAt(i);
			game.alphabet.pushBack(letter);
			if (game.playerSlots.contains(letter))	letter.enable();
		}
	}
	
	// return a String-encoded version of the slots.
	public String toString () {
		String s = "";
		for (int i=0; i<slots.size(); i++) {
			CGLetterSlot nextSlot = slots.elementAt(i);
			String nextLetter = nextSlot.getLetter().toString();
			if (nextLetter.equals(""))	s = s + CGProtocol.LETTERSLOT_EMPTY;
			else						s = s + nextLetter;
		}
		return s;
	}
	
	// set slots according to given String-encoded data.
	// returns FALSE on invalid data, otherwise TRUE.
	public boolean set(String l) {
		if (l == null)	    								return false;
		if ( (l.length() == 0) || (l.length() != MAXSLOTS) )return false;
		
		// set the slots according to data.
		for (int i=0; i<l.length(); i++) {
			String nextLetter = l.substring(i, i+1);
			if (nextLetter.equals(CGProtocol.LETTERSLOT_EMPTY))
				slots.elementAt(i).setLetter(new CGLetter());
			else
				slots.elementAt(i).setLetter(new CGLetter(nextLetter));
		}
		return true;
	}
}
