


// the moves, which a player can make during his/her turn,
// are descibed by this class for the purpose of transmitting
// them between the player and the opponent ends.
public class CGMove {
	
	// each move is described by the contents of a
	// subset of the following fields.
	final private CGLetterSlot source;
	final private CGLetterSlot target;
	final private String type; 
	final private int score;
	final private String alphabet;
	
	// the parent game.
	final private CrosswordGame game;

	
	
	// the constructor for wordfix and undo wordfix moves.
	public CGMove(final CrosswordGame g, final String y, final int points) {
		game = g;
		source = null;
		target = null;
		score = points;
		alphabet = "";
		
		if (y.equals(CGProtocol.MOVE_TYPES_FIX))			type = CGProtocol.MOVE_TYPES_FIX;
		else if (y.equals(CGProtocol.MOVE_TYPES_UNDOFIX))	type = CGProtocol.MOVE_TYPES_UNDOFIX;
		else												type = CGProtocol.MOVE_TYPES_INVALID;
	}

	// the constructor for letter-moving moves.
	public CGMove(final CrosswordGame g, final CGLetterSlot s, final CGLetterSlot t, final String y) {
		game = g;
		type = y;
		source = s;
		target = t;
		score = 0;
		alphabet = "";
	}

	// the structure of s String-encoded letter-moving move is
	// TQTQXQYTXTY
	// `----------- T  type
	//  `---------- Q  queued letter index
	//   `--------- T  target letter index
	//     `------- QX queued cell x coordinate
	//       `----- QY queued cell y coordinate
	//          `-- TX target cell x coordinate
	//            ` TY target cell y coordinate
	//
	// not all fields are used simultaneously.
	// unused fields are padded with zeroes.
	// the type is the only field that is always present.
	// only one queued and one target are transmitted at a time.
	// when a queued cell is transmitted with a target letter,
	// the padding is extended to the end.
	//
	// the structure of a String-encoded cancel is
	// T
	// - T type
	//
	// the structure of a String-encoded letter swap is
	// TA...
	// `----- T		type
	//  `---- A...	a CGLetterSlots+1 to CGSettings.ALPHABETMAXIMUM length Alphabet.
	//
	// the structure of a String-encoded wordfix is
	// TN...
	// `----- T 	type
	//  `---- N...	an arbitrarily long integer
	//				representing score associated with the move.
	//
	// the structure of a String-encoded undo wordfix is
	// TN...
	// `----- T 	type
	//  `---- N...	an arbitrarily long integer
	//				representing score associated with the move.
	//
	// NOTE: encode() is NOT an equivalent of a "toString()"!
	// it should NOT be used to make printouts of Moves, because the String 
	// representation encode() returns depends on locale.
	// encode() should only be used in its name sake, encoding moves for transfer.
	public String encode() {
		String s = type;
		
		// a crossword-to-crossword move.
		if (type.equals(CGProtocol.MOVE_TYPES_C2C)) {
			final Integer qx = new Integer(game.crossword.xOf((CGCell) source));
			final Integer qy = new Integer(game.crossword.yOf((CGCell) source));
			final Integer tx = new Integer(game.crossword.xOf((CGCell) target));
			final Integer ty = new Integer(game.crossword.yOf((CGCell) target));
			final String sqx = (qx <= 9) ? "0" + qx : String.valueOf(qx);
			final String sqy = (qy <= 9) ? "0" + qy : String.valueOf(qy);
			final String stx = (tx <= 9) ? "0" + tx : String.valueOf(tx);
			final String sty = (ty <= 9) ? "0" + ty : String.valueOf(ty);
			
			s = s + "0" + "0";		// required padding.
			s = s + sqx + sqy + stx + sty;
		}
		
		// a crossword-to-slot move.
		else if (type.equals(CGProtocol.MOVE_TYPES_C2O)) {
			final Integer qx = new Integer(game.crossword.xOf((CGCell) source));
			final Integer qy = new Integer(game.crossword.yOf((CGCell) source));
			final Integer it = new Integer(game.playerSlots.indexOf(target));
			final String sx = (qx <= 9) ? "0" + qx : String.valueOf(qx);
			final String sy = (qy <= 9) ? "0" + qy : String.valueOf(qy);
			
			s = s + "0";			// required padding.
			s = s + it.toString();
			s = s + sx + sy;
			s = s + "0000";			// required padding.
		}
		
		// a slot-to-crossword move.
		else if (type.equals(CGProtocol.MOVE_TYPES_O2C)) {
			final Integer iq = new Integer(game.playerSlots.indexOf(source));
			final Integer tx = new Integer(game.crossword.xOf((CGCell) target));
			final Integer ty = new Integer(game.crossword.yOf((CGCell) target));
			final String sx = (tx <= 9) ? "0" + tx : String.valueOf(tx);
			final String sy = (ty <= 9) ? "0" + ty : String.valueOf(ty);

			s = s + iq.toString();
			s = s + "0" + "0000";	// required padding.
			s = s + sx + sy;
		}
		
		// a slot-to-slot move.
		else if (type.equals(CGProtocol.MOVE_TYPES_O2O)) {
			final Integer iq = new Integer(game.playerSlots.indexOf(source));
			final Integer it = new Integer(game.playerSlots.indexOf(target));
			
			s = s + iq.toString();
			s = s + it.toString();
		}
		
		// a letter swap.
		else if (type.equals(CGProtocol.MOVE_TYPES_SWAP)) {
			s = type + game.alphabet.toString();
		}
		
		// a cancel.
		else if (type.equals(CGProtocol.MOVE_TYPES_CANCEL)) {
			s = type;
		}
		
		// a wordfix.
		else if (type.equals(CGProtocol.MOVE_TYPES_FIX)) {
			s = type + Integer.toString(score);
		}
		
		// an undo wordfix.
		else if (type.equals(CGProtocol.MOVE_TYPES_UNDOFIX)) {
			s = type + Integer.toString(score);
		}
		
		return s;
	}

	// build an executable move from a String-encoded move.
	public CGMove(final CrosswordGame g, final String move) {
		game = g;
		
		type = getMoveType(move);
		
		// a crossword-to-crossword move.
		if (type.equals(CGProtocol.MOVE_TYPES_C2C)) {
			
			int qx = getQueuedCellX(move);
			int qy = getQueuedCellY(move);
			source = game.crossword.getCrosswordCell(qx, qy);

			int tx = getTargetCellX(move);
			int ty = getTargetCellY(move);
			target = game.crossword.getCrosswordCell(tx, ty);
			
			score = 0;
			alphabet = "";
		}
		
		// a crossword-to-slot move.
		else if (type.equals(CGProtocol.MOVE_TYPES_C2O)) {

			int qx = getQueuedCellX(move);
			int qy = getQueuedCellY(move);
			source = game.crossword.getCrosswordCell(qx, qy);
			
			int ti = getTargetO(move);
			target = game.opponentSlots.elementAt(ti);

			score = 0;
			alphabet = "";
		}
		
		// a slot-to-crossword move.
		else if (type.equals(CGProtocol.MOVE_TYPES_O2C)) {

			int qi = getQueuedO(move);
			source = game.opponentSlots.elementAt(qi); 
			
			int tx = getTargetCellX(move);
			int ty = getTargetCellY(move);
			target = game.crossword.getCrosswordCell(tx, ty);

			score = 0;
			alphabet = "";
		}
		
		// a slot-to-slot move.
		else if (type.equals(CGProtocol.MOVE_TYPES_O2O)) {

			int qi = getQueuedO(move);
			source = game.opponentSlots.elementAt(qi); 

			int ti = getTargetO(move);
			target = game.opponentSlots.elementAt(ti);
		
			score = 0;
			alphabet = "";
		}

		// a letter swap.
		else if (type.equals(CGProtocol.MOVE_TYPES_SWAP)) {
			alphabet = getAlphabet(move);
			source = null;
			target = null;
			score = 0;
		}
		
		// a cancel.
		else if (type.equals(CGProtocol.MOVE_TYPES_CANCEL))  {
			source = null;
			target = null;
			score = 0;
			alphabet = "";
		}
		
		// a wordfix.
		else if (type.equals(CGProtocol.MOVE_TYPES_FIX)) {
			score = getScore(move);
			source = null;
			target = null;
			alphabet = "";
		}
		
		// an undo wordfix.
		else if (type.equals(CGProtocol.MOVE_TYPES_UNDOFIX)) {
			score = getScore(move);
			source = null;
			target = null;
			alphabet = "";
		}
		
		// invalid moves.
		else {
			source = null;
			target = null;
			score = 0;
			alphabet = "";
		}
	}

	// execute the move.
	public boolean execute() {

		game.queue.clear();

		// a crossword-to-crossword move.
		if (type.equals(CGProtocol.MOVE_TYPES_C2C)) {
			game.queue.set(source);
			game.queue.swap(target);
			return true;
		}

		// a crossword-to-slot move.
		else if (type.equals(CGProtocol.MOVE_TYPES_C2O)) {
			game.queue.set(source);
			game.queue.swap(target);
			return true;
		}

		// a slot-to-crossword move.
		if (type.equals(CGProtocol.MOVE_TYPES_O2C)) {
			game.queue.set(source);
			game.queue.swap(target);
			return true;
		}

		// a slot-to-slot move.
		else if (type.equals(CGProtocol.MOVE_TYPES_O2O))  {
			game.queue.set(source);
			game.queue.swap(target);
			return true;
		}
		
		// a letter swap.
		else if (type.equals(CGProtocol.MOVE_TYPES_SWAP)) {
    		game.crossword.cancelOpponentWords();
    		game.alphabet.set(alphabet);
    		game.opponentSlots.initialize();
    		game.opponentSlots.disable();
			return true;
		}
		
		// a cancel.
		else if (type.equals(CGProtocol.MOVE_TYPES_CANCEL)) {
			game.crossword.cancelOpponentWords();
			return true;
		}
		
		// a wordfix.
		else if (type.equals(CGProtocol.MOVE_TYPES_FIX)) {
			game.crossword.fixWords();
			game.opponentSlots.fill();
			game.scoreTable.increaseOpponentScore(score);
			game.scoreTable.undoPts = score;
			game.scoreTable.setAlphabetCounter();
    		game.menus.enableUndo();
			return true;
		}
		
		// an undo wordfix.
		else if (type.equals(CGProtocol.MOVE_TYPES_UNDOFIX)) {
			game.crossword.cancelOpponentWords();
			game.playerSlots.undo();
			game.crossword.undoPlayer();
			game.scoreTable.undoPlayerScore();
    		game.scoreTable.setAlphabetCounter();
    		game.menus.disableUndo();
			return true;
		}
		
		// unknown type (should never be reached).
		else	return false;
	}

	
	
    // validate given String-encoded move.
    public static boolean isValidMove(String move) {
    	
    	// there are no empty moves.
    	if (move == null)		return false;
		if (move.length() < 1)	return false;
		
		String type = getMoveType(move);
		
		// a cancel is the most primitive of moves.
    	if (type.equals(CGProtocol.MOVE_TYPES_CANCEL) ) {
    		return true;
    	}
    	
		// a wordfix.
    	if (type.equals(CGProtocol.MOVE_TYPES_FIX)) {
    		try {
    			Integer.parseInt(move.substring(1));
    		}
    		catch (NumberFormatException ex) {
    			return false;
    		}
    		
    		return true;
    	}
    	
		// undo wordfix.
    	if (type.equals(CGProtocol.MOVE_TYPES_UNDOFIX)) {
    		try {
    			Integer.parseInt(move.substring(1));
    		}
    		catch (NumberFormatException ex) {
    			return false;
    		}
    		
    		return true;
    	}
    	
		// the only valid moves described by a String shorter than 3 characters
    	// have been excluded by the previous clauses.
		if (move.length() < 3)return false;

    	// a swap.
    	// since swapping is disabled when there are no more letters in the Alphabet,
		// the new Alphabet has CGLetterSlots+1 letters at shortest. so, as long as
		// CGLetterSlots > 0, the "at least three characters" clause above will be filled.
		// the upper limit is tested by the next condition.
		if (type.equals(CGProtocol.MOVE_TYPES_SWAP) && move.length()-1 > CGSettings.ALPHABETMAXIMUM)
			return false;
		
		// moves which involve crossword cells are
		// always padded to a length of 11 characters.
		if (
			(type.equals(CGProtocol.MOVE_TYPES_O2C)
			|| type.equals(CGProtocol.MOVE_TYPES_C2O)
			|| type.equals(CGProtocol.MOVE_TYPES_C2C))
			&& (move.length() != 11)
		)	return false;

		return true;
    }
    
    // pick the move type from a String-encoded move.
    public static String getMoveType(String move) {
    	if (move == null)		return CGProtocol.MOVE_TYPES_INVALID;
    	if (move.length() < 1)	return CGProtocol.MOVE_TYPES_INVALID;
    	if (move.substring(0,1).equals(CGProtocol.MOVE_TYPES_C2C))			return CGProtocol.MOVE_TYPES_C2C;
    	else if (move.substring(0,1).equals(CGProtocol.MOVE_TYPES_C2O))		return CGProtocol.MOVE_TYPES_C2O;
    	else if (move.substring(0,1).equals(CGProtocol.MOVE_TYPES_O2C))		return CGProtocol.MOVE_TYPES_O2C;
    	else if (move.substring(0,1).equals(CGProtocol.MOVE_TYPES_O2O))		return CGProtocol.MOVE_TYPES_O2O;
    	else if (move.substring(0,1).equals(CGProtocol.MOVE_TYPES_SWAP))	return CGProtocol.MOVE_TYPES_SWAP;
    	else if (move.substring(0,1).equals(CGProtocol.MOVE_TYPES_CANCEL))	return CGProtocol.MOVE_TYPES_CANCEL;
    	else if (move.substring(0,1).equals(CGProtocol.MOVE_TYPES_FIX))		return CGProtocol.MOVE_TYPES_FIX;
    	else if (move.substring(0,1).equals(CGProtocol.MOVE_TYPES_UNDOFIX))	return CGProtocol.MOVE_TYPES_UNDOFIX;
    	else																return CGProtocol.MOVE_TYPES_INVALID;
    }
    
    // pick type-specific data from a String-encoded move.
    private String getAlphabet(String move) {
    	return move.substring(1);
    }
    private int getScore(String move) {
    	return new Integer(move.substring(1)).intValue();
    }
    private int getQueuedO(String move) {
    	return new Integer(move.substring(1,2)).intValue();
    }
    private int getTargetO(String move) {
    	return new Integer(move.substring(2,3)).intValue();
    }
    private int getQueuedCellX(String move) {
    	return new Integer(move.substring(3,5)).intValue();
    }
    private int getQueuedCellY(String move) {
    	return new Integer(move.substring(5,7)).intValue();
    }
    private int getTargetCellX(String move) {
    	return new Integer(move.substring(7,9)).intValue();
    }
    private int getTargetCellY(String move) {
    	return new Integer(move.substring(9,11)).intValue();
    }
}
