import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Spinner;



// the score table keeps and shows both players' score
// and the number of available alphabet.
public class CGScoreTable extends Composite {

	// labels for the player's score and the opponent's score.
	final private static String PLAYERSCORE		= "Omat pisteet:";
	final private static String OPPONENTSCORE	= "Vastustajan pisteet:";

	// label and tooltip for the Alphabet counter.
	final private static String ALPHABETCOUNTER_LABEL		= "Kirjaimia:";
	final private static String ALPHABETCOUNTER_TOOLTIP		= "Montako kirjainta on jljell.";
	
	// resort to zero as the minimum score because Spinner,
	// which used to handle the player's score, cannot handle negative values.
	// theoretically it's possible to end up with a negative score at the end of game,
	// but it's such a reach that this limit shouldn't pose any practical problems.
	final private static int MINIMUM	= 0;
	
    // a reasonable maximum for a score. 
	// a theoretical maximum would be reaced in a game where only
	// one of the players would lay down letters (the other would
	// simply pass all turns). the player would fill all the crossword 
	// cells, thus being awarded not only one point per each cell of 
	// the grid, but also additional points for the crossing words.
	// in practice the score is always much lower, since there are two
	// players splitting the score (i.e. the opponent also lays down
	// letters) and only few of the crossing words are accepted for
	// scoring.
	// CGSubmitDialog needs to see this.
    final public static int MAXIMUM	= 2 * CGCrossword.ROWS * CGCrossword.COLUMNS;
    
	// the player's score is kept in a Spinner.
	final private Spinner playerScore;
	
	// the number of remaining Alphabet is kept in a Label.
	final private AlphabetCounter alphabetCounter;
	
	// the opponent's score is kept in a Label.
	final private Label opponentScore;
	
    // the last recorded score for the player.
    private int lastScore;
    
	// undo points.
    // CGMenu and CGMove need to see this.
	public int undoPts;
	
	
	
	// the constructor's parameters are the parent window and game.
	final CrosswordGame game;
	public CGScoreTable(final Shell shell, final CrosswordGame g) {
		super(shell, SWT.NONE);
		GridLayout gl = new GridLayout(5, false);
		GridData fill = new GridData(GridData.FILL_HORIZONTAL);
		this.setLayout(gl);
		this.setLayoutData(fill);

		game = g;

		// the player's score.
		new Label(this, SWT.NONE).setText(PLAYERSCORE);
	    playerScore = new Spinner(this, SWT.BORDER);
	    playerScore.setMinimum(MINIMUM);
	    playerScore.setMaximum(MAXIMUM);
	    playerScore.setIncrement(1);
	    playerScore.setPageIncrement(10);

	    // the alphabet counter.
	    alphabetCounter = new AlphabetCounter(this);
	    GridData center = new GridData(GridData.CENTER);
	    center.widthHint = 160;
	    alphabetCounter.setLayoutData(center);

	    // the opponent's score.
		Label vp = new Label(this, SWT.NONE);
		vp.setText(OPPONENTSCORE);
		fill = new GridData(GridData.FILL_HORIZONTAL | GridData.HORIZONTAL_ALIGN_END);
		vp.setLayoutData(fill);
		fill = new GridData(GridData.HORIZONTAL_ALIGN_END);
		fill.widthHint = 21;
		opponentScore = new Label(this, SWT.NONE);
		opponentScore.setLayoutData(fill);
		
		initialize();
	}

	// set all counters to their initial values.
	// CrosswordGame needs to see this.
	public void initialize() {
		playerScore.setSelection(0);
		setAlphabetCounter();
		opponentScore.setText("0");
		alphabetCounter.notBlinked = true;
		lastScore = 0;
		undoPts = 0;
	}
	
	// encapsulation for the alphabet counter.
	// CGClient, CGMenu and CGMove need to see this.
	public void setAlphabetCounter() {
		alphabetCounter.set();
	}

	// encapsulation for last recorded score.
	// CGActionButtonSet needs to see these.
	public int getLastScore() {
		return lastScore;
	}
	public void recordScore(final int ptsForMove) {
		lastScore = game.scoreTable.getPlayerScore();
		undoPts = ptsForMove;
	}

	// encapsulation for the player's score.
	// CGActionButtonSet, CGClient, CGMove needs to see these.
	public int getPlayerScore () {
		return playerScore.getSelection();
	}
	public void increasePlayerScore (int points) {
		playerScore.setSelection(playerScore.getSelection() + points);
	}
	public void undoPlayerScore() {
		playerScore.setSelection(playerScore.getSelection() -undoPts);
		recordScore(0);
	}
	// 'restore' is not the same as 'undo'!
	// 'undo' takes away points given for a fixed move, whereas 'restore' restores 
	// the score back to what it was at the beginning of the current, yet unfixed move.
	public void restoreScore() {
		playerScore.setSelection(lastScore);
	}
	public String playerScoreToString() {
		return Integer.toString(playerScore.getSelection());
	}
	public boolean setPlayerScore(String score) {
		try {
			int newScore = Integer.parseInt(score);
			playerScore.setSelection(newScore);
			return true;
		}
		catch (NumberFormatException ex) {
			return false;
		}
	}

	// encapsulation for the opponent's score.
	// CGClient, CGMenu and CGMove need to see these.
	public void increaseOpponentScore (int points) {
		opponentScore.setText(
			new Integer (
				new Integer (
					opponentScore.getText()
				).intValue() + points
			).toString()
		);
	}
	public String opponentScoreToString() {
		return opponentScore.getText();
	}
	public boolean setOpponentScore(String score) {
		try {
			Integer.parseInt(score);
			opponentScore.setText(score);
			return true;
		}
		catch (NumberFormatException ex) {
			return false;
		}
	}

	// disable/enable the score table.
	public void disable() {
		playerScore.setEnabled(false);
	}
	public void enable() {
		playerScore.setEnabled(true);
	}
	
	
	
	// the alphabet counter.
	private class AlphabetCounter extends Composite {

		// the limit (in letters) for the alphabet under which 
		// the number of letters left will be highlighted by blinking.
		final private static int BLINKLIMIT = CGLetterSlots.MAXSLOTS;

		// the number od letters left.
		final private Label alphabetCount;
		
		// whether the counter has been blinked yet.
		private boolean notBlinked = true;
		
		
		
		// the constructor.
		public AlphabetCounter(final Composite parent) {
			super(parent, SWT.NONE);
			GridData fill = new GridData(GridData.FILL_HORIZONTAL | GridData.HORIZONTAL_ALIGN_END);
			this.setLayout(new GridLayout(2, false));
			this.setLayoutData(fill);

			// the label for the counter.
			Label alphabetLabel = new Label(this, SWT.NONE);
			alphabetLabel.setText(CGScoreTable.ALPHABETCOUNTER_LABEL);
			alphabetLabel.setToolTipText(CGScoreTable.ALPHABETCOUNTER_TOOLTIP);
		    GridData align = new GridData(GridData.FILL_HORIZONTAL | GridData.HORIZONTAL_ALIGN_END);
			alphabetLabel.setLayoutData(align);

			// the counter.
			alphabetCount = new Label(this, SWT.NONE);
			align = new GridData(GridData.HORIZONTAL_ALIGN_BEGINNING);
			align.widthHint = 50;
			alphabetCount.setLayoutData(align);
			alphabetCount.setText("0");
			alphabetCount.setToolTipText(CGScoreTable.ALPHABETCOUNTER_TOOLTIP);
		}
		
		// 'update' would be a more descriptive name, but it's reserved by the parent class.
		private void set() {
			int count  = game.alphabet.size();
			alphabetCount.setText(new Integer(count).toString());
			
			// blink four times to signal diminishing alphabet.
			if ( (count < BLINKLIMIT) && notBlinked ) {
				final Color red = Display.getDefault().getSystemColor(SWT.COLOR_RED);
				final Color original = alphabetCount.getBackground();

				for (int i=0; i<4; i++) {
					Display.getDefault().timerExec(400*i, new Runnable() {
						public void run() {
							alphabetCount.setBackground(red);
						}
					});
					Display.getDefault().timerExec(200+400*i, new Runnable() {
						public void run() {
							alphabetCount.setBackground(original);
						}
					});
				}
				notBlinked = false;
			}
		}
	}
}
