import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.net.InetAddress;
import java.net.UnknownHostException;



// the game settings.
public class CGSettings {
	
	// the settings file name, the number of setting items.
	final private static String SETTINGSFILENAME	= "settings.txt";
	final private static int	SETTINGS			= 2;

	// the items in the settings file. the settings file is made up of lines consisting of
	// item=value
	// where item is a constant set below here, and value is the saved setting of the item.
	final private static String PLAYERITEMNAME 		= "player";
	final private static String OPPONENTIPITEMNAME	= "opponent";

	// the pre-defined setting values for the setting file.
	// CGMenu needs the player values,
	// CGAddressDialog needs the opponent IP.
	final public static int PLAYER0 = 0;	// a non-valid player
	final public static int PLAYER1 = 1;
	final public static int PLAYER2 = 2;
	final public static String DEFAULT_OPPONENT_IP	= "127.0.0.1";

	// the alphabet file name, the maximum number of letters to be read.
	// ScoreTable needs to see the maximum.
	final private static String ALPHABETFILENAME	= "alphabet.txt";
	final public static int 	ALPHABETMAXIMUM		= CGCrossword.COLUMNS * CGCrossword.ROWS;
	
	// the currently selected player (i.e. "who am I").
	private int player;

	// the opponent's IP address setting.
	private String opponentIP;

	// the alphabet setting.
	private String alphabet;

	
	
	// the constructor initially sets all items to their defaults.
	// loading saved settings needs to be done separately by invoking load().
	public CGSettings() {
		setPlayer(PLAYER1);
		alphabet = 
			"AAAAAAAAAAAA" +
			"TTTTTTTTTTT" +
			"IIIIIIIIIII" +
			"EEEEEEE" +
			"UUUUUU" +
			"KKKKKK" +
			"SSSSSS" +
			"LLLLLL" + 
			"OOOOO" +
			"NNNNN" +
			"RRRR" +
			"" +
			"MMM" +
			"PPP" +
			"HH" +
			"VV" +
			"YY" +
			"JJ" +
			"D" +
			""
		;
		
		setOpponentIP(DEFAULT_OPPONENT_IP);
	}
	
	// opponent's IP address setting's encapsulation.
	// CGConnection and CGAddressDialog need to see these.
	public String getOpponentIP() {
		return opponentIP;
	}
	public boolean setOpponentIP(String s) {
		try {
			InetAddress.getByName(s);
			opponentIP = s;
			return true;
		}
		catch (UnknownHostException ex) {
			opponentIP = DEFAULT_OPPONENT_IP;
			return false;
		}
	}
	
	// the alphabet setting's encapsulation.
	// CGAlphabet needs to see this.
	public String getAlphabet() {
		return alphabet;
	}	

	// set the current player.
	// returns FALSE if the given player isn't a valid value
	// (i.e. neither player 1 nor player 2), otherwise returns TRUE.
	// CGMenu needs to see this.
	public boolean setPlayer(final int p) {
		if ( (p != PLAYER1) && (p != PLAYER2) )	return false;
		
		player = p;
		return true;
	}

	// get the current player, as seen by the settings.
	// CGMenu needs to see this.
	public int getPlayer() {
		return player;
	}
	
	// settings saver.
	// CGAddressDialog and CrosswordGame need to see this.
	public boolean save() {
	    try {
	    	final PrintStream out = new PrintStream(new FileOutputStream(SETTINGSFILENAME));
	    	out.println(PLAYERITEMNAME + "=" + player);
	    	out.println(OPPONENTIPITEMNAME + "=" + opponentIP);
	    	out.close();
	    	return true;
	      
	    }

	    catch (IOException ex) {
	    	return false;
	    }
	}

	// settings loader.
	// CrosswordGame needs to see this.
	public boolean load() {
		try {
			
			// number of setting items read.
			int readCount = 0;
			
			// open the file.
			final BufferedReader in = new BufferedReader(new FileReader(SETTINGSFILENAME));

			// initialize the storage buffer.
			StringBuffer settings[] = new StringBuffer[SETTINGS];
			for (int i=0; i<SETTINGS; i++)	settings[i] = new StringBuffer();
			
			// read until EOF but only a preset amount of setting items.
			String word = in.readLine();
			while (word != null) {
				settings[readCount].append(word);
				readCount++;
				if ( !(readCount < SETTINGS) )	break;
				word = in.readLine();
			}

			// close the file.
			in.close();
			
			// parse the read data.
			return parseSettings(settings);
		}
		
		// fail silently on read errors.
		catch (FileNotFoundException ex) {
			return false;
		}
		catch (IOException ex) {
			return false;
		}
	}
	
	// the method for parsing an array of settings into settings fields' data.
	// returns TRUE if at least one setting was parsed successfully, otherwise FALSE.
	private boolean parseSettings(StringBuffer settings[]) {

		// if the settings array is empty, fail immediately.
		if ( !(settings.length > 0) )	return false;
			
		// number of successfully parsed setting items.
		int pSuccess = 0;
		
		// go through the raw settings array.
		for (int i=0; i<settings.length ; i++) {

			// find the item name and data separator.
			int equalsAt = settings[i].indexOf("=");

			// if an item name and data
			// (i.e. something following the "=")
			// was found from the line.
			if ( (equalsAt > 0) && settings[i].length() > equalsAt+1 ) {
				
				// grab the setting key and item as Strings.
				String item = settings[i].substring(0,equalsAt);
				String setting = settings[i].substring(equalsAt+1);

				// the player setting item.
				if (item.equals(PLAYERITEMNAME)) {
					
					int p;
					
					// validate the setting.
					// it needs to be an integer.
					try {
						p = Integer.parseInt(setting);
					}
					catch (NumberFormatException ex) {
						p = PLAYER0;
					}
					
					// if a valid player setting was found.
					if (p != PLAYER0) {
						
						// set the actual field according to read data.
						setPlayer(p);

						// increase number of successfully parsed setting items.
						pSuccess++;
					}
					// no else: skip an invalid setting,
					// i.e. do not change current value of player
					// if an invalid setting was read.
				}
				
				else if (item.equals(OPPONENTIPITEMNAME))
					if (setOpponentIP(setting)) pSuccess++;
				
				// other settings would be picked up here.
			}
			// no else: invalid lines are silently skipped.
		}
		
		// if at least one setting was successfully parsed,
		// return TRUE for success.
		// otherwise return FALSE to signal failure.
		return (pSuccess > 0);
	}

	// alphabet loader.
	// returns TRUE if at least one setting is read successfully, otherwise FALSE.
	public boolean loadAlphabet() {
		String readAlphabet = "";
		try {
			
			// open the alphabet file.
			final FileInputStream in = new FileInputStream(ALPHABETFILENAME);

			// temporary storage for input.
			int c;
			
			// number of letters read.
			int read = 0;
			
			// read until EOF.
			do {
				
				// read, skipping all but letters.
				c = in.read();
				if (Character.isLetter((char) c)) {
					c = Character.toUpperCase( (char)c );
					readAlphabet = readAlphabet + (char)c;
					read++;
					if (read > ALPHABETMAXIMUM)	break;
				}
			} while (c != -1);
			// read until EOF.
			// http://java.sun.com/j2se/1.4.2/docs/api/java/io/FileInputStream.html#read()

			// close the file.
			in.close();
			
			// if no valid letters were read.
			if (readAlphabet.length() == 0)		return false;

			alphabet = readAlphabet;
			return true;
		}
		
		// fail silently on read errors.
		catch (FileNotFoundException ex) {
			return false;
		}
		catch (IOException ex) {
			return false;
		}
	}
}
